# -*- coding: utf-8 -*-

"""
    The training pixel metrics module
    ======================

    Use it to compute different metrics during training.
    Available metrics:
        - Confusion matrix
        - Intersection-over-Union
"""

import numpy as np


def compute_metrics(
    pred: np.ndarray, label: np.ndarray, loss: float, classes: list
) -> dict:
    """
    Compute the metrics between a prediction and a label mask.
    :param pred: The prediction made by the network.
    :param label: The mask of the input image.
    :param loss: The loss the the current batch.
    :param classes: The classes names involved during the experiment.
    :return metrics: The computed metrics.
    """
    metrics = {}
    metrics["matrix"] = confusion_matrix(pred, label, classes)
    metrics["loss"] = loss
    return metrics


def update_metrics(metrics: dict, batch_metrics: dict) -> dict:
    """
    Add batch metrics to the global metrics.
    :param metrics: The global epoch metrics.
    :param batch_metrics: The current batch metrics.
    :return metrics: The updated global metrics.
    """
    for i in range(metrics["matrix"].shape[0]):
        for j in range(metrics["matrix"].shape[1]):
            metrics["matrix"][i][j] += batch_metrics["matrix"][i][j]
    metrics["loss"] += batch_metrics["loss"]
    return metrics


def confusion_matrix(pred: np.ndarray, label: np.ndarray, classes: list) -> np.array:
    """
    Get the confusion matrix between the prediction and the given label.
    :param pred: The prediction made by the network.
    :param label: The mask of the input image.
    :params classes: The classes names involved during the experiment.
    :return confusion_matrix: The computed confusion matrix.
    """
    size = len(classes)
    confusion_matrix = np.zeros((size, size))
    for i in range(size):
        for j in range(size):
            bin_label = label == i
            bin_pred = pred == j
            confusion_matrix[j, i] = (bin_pred * bin_label).sum()
    return confusion_matrix


def iou(confusion_matrix: np.ndarray, channel: str) -> float:
    """
    Get the Intersection-over-Union values between the prediction and the
    given label. It returns one value for the given class.
    :param confusion_matrix: The confusion matrix obtained between the label
                             and prediction masks.
    :param channel: The name of the current class.
    :return: The computed Intersection-over-Union value.
    """
    true_positives = confusion_matrix[channel, channel]
    tpfn = np.sum(confusion_matrix[:, channel])
    tpfp = np.sum(confusion_matrix[channel, :])
    if true_positives == 0:
        return 0
    elif tpfn + tpfp == true_positives:
        return 0
    return true_positives / (tpfn + tpfp - true_positives)



# -*- coding: utf-8 -*-

import argparse
import json
from pathlib import Path

import gradio as gr
import numpy as np
from PIL import Image, ImageDraw

from hugging_face.config import parse_configurations
from hugging_face.tools import UFCNModel

# Parse the CLI arguments
parser = argparse.ArgumentParser(description="UFCN HuggingFace app")
parser.add_argument(
    "--config",
    type=Path,
    required=True,
    help="The YAML-formatted configuration file of the Hugging Face app",
)

# Get the application's public mode (local or with sharing)
parser.add_argument(
    "--public",
    action="store_true",
    help="Generate a shareable link to your application",
)

# Parse arguments
args = parser.parse_args()

# Load the config
config = parse_configurations(args.config)

# Check that the paths of the examples are valid
for example in config["examples"]:
    assert Path.exists(
        Path(example)
    ), f"The path of the image '{example}' does not exist."

# Cached models, maps model_name to UFCNModel object
MODELS = {
    model["model_name"]: UFCNModel(
        name=model["model_name"],
        colors=model["classes_colors"],
        title=model["title"],
        description=model["description"],
    )
    for model in config["models"]
}

# Create a list of models name
models_name = list(MODELS)


def load_model(model_name) -> UFCNModel:
    """
    Retrieve the model, and load its parameters/files if it wasn't done before.

    :param model_name: The name of the selected model
    :return: The UFCNModel instance selected
    """
    assert model_name in MODELS
    model = MODELS[model_name]
    # Load the model's files if it wasn't done before
    if not model.loaded:
        model.load()
    return model


def query_image(model_name: gr.Dropdown, image: gr.Image) -> list([Image, json]):
    """
    Loads a model and draws the predicted polygons with the color provided by the model on an image

    :param model: A model selected in dropdown
    :param image: An image to predict
    :return: Image and dict, an image with the predictions and a
        dictionary mapping an object idx (starting from 1) to a dictionary describing the detected object:
        - `polygon` key : list, the coordinates of the points of the polygon,
        - `confidence` key : float, confidence of the model,
        - `channel` key : str, the name of the predicted class.
    """

    # Load the model and get its classes, classes_colors and the model
    ufcn_model = load_model(model_name)

    # Make a prediction with the model
    detected_polygons, probabilities, mask, overlap = ufcn_model.model.predict(
        input_image=image, raw_output=True, mask_output=False, overlap_output=False
    )

    # Load image
    image = Image.fromarray(image)

    # Make a copy of the image to keep the source and also to be able to use Pillow's blend method
    img2 = image.copy()

    # Initialize the dictionary which will display the json on the application
    predict = []

    # Create the polygons on the copy of the image for each class with the corresponding color
    # We do not draw polygons of the background channel (channel 0)
    for channel in range(1, ufcn_model.num_channels):
        for i, polygon in enumerate(detected_polygons[channel]):
            # Draw the polygons on the image copy.
            # Loop through the class_colors list (channel 1 has color 0)
            ImageDraw.Draw(img2).polygon(
                polygon["polygon"], fill=ufcn_model.colors[channel - 1]
            )

            # Build the dictionary
            # Add an index to dictionary keys to differentiate predictions of the same class
            predict.append(
                {
                    # The list of coordinates of the points of the polygon.
                    # Cast to list of np.int32 to make it JSON-serializable
                    "polygon": np.asarray(polygon["polygon"], dtype=np.int32).tolist(),
                    # Confidence that the model predicts the polygon in the right place
                    "confidence": polygon["confidence"],
                    # The channel on which the polygon is predicted
                    "channel": ufcn_model.classes[channel],
                }
            )

    # Return the blend of the images and the dictionary formatted in json
    return Image.blend(image, img2, 0.5), json.dumps(predict, indent=2)


def update_model(model_name: gr.Dropdown) -> str:
    """
    Update the model title to the title of the current model

    :param model_name: The name of the selected model
    :return: A new title
    """
    return f"## {MODELS[model_name].title}", MODELS[model_name].description


with gr.Blocks() as process_image:
    # Create app title
    gr.Markdown(f"# {config['title']}")

    # Create app description
    gr.Markdown(config["description"])

    # Create dropdown button
    model_name = gr.Dropdown(models_name, value=models_name[0], label="Models")

    # get models
    selected_model: UFCNModel = MODELS[model_name.value]

    # Create model title
    model_title = gr.Markdown(f"## {selected_model.title}")

    # Create model description
    model_description = gr.Markdown(selected_model.description)

    # Change model title and description when the model_id is update
    model_name.change(update_model, model_name, [model_title, model_description])

    # Create a first row of blocks
    with gr.Row():
        # Create a column on the left
        with gr.Column():
            # Generates an image that can be uploaded by a user
            image = gr.Image()

            # Create a row under the image
            with gr.Row():
                # Generate a button to clear the inputs and outputs
                clear_button = gr.Button("Clear", variant="secondary")

                # Generates a button to submit the prediction
                submit_button = gr.Button("Submit", variant="primary")

            # Create a row under the buttons
            with gr.Row():
                # Generate example images that can be used as input image for every model
                gr.Examples(config["examples"], inputs=image)

        # Create a column on the right
        with gr.Column():
            with gr.Row():
                # Generates an output image that does not support upload
                image_output = gr.Image(interactive=False)

            # Create a row under the predicted image
            with gr.Row():
                # Create a column so that the JSON output doesn't take the full size of the page
                with gr.Column():
                    # # Create a collapsible region
                    with gr.Accordion("JSON"):
                        # Generates a json with the model predictions
                        json_output = gr.JSON()

    # Clear button: set default values to inputs and output objects
    clear_button.click(
        lambda: (None, None, None),
        inputs=[],
        outputs=[image, image_output, json_output],
    )

    # Create the button to submit the prediction
    submit_button.click(
        query_image, inputs=[model_name, image], outputs=[image_output, json_output]
    )

# Launch the application with the public mode (True or False)
process_image.launch(share=args.public)



# -*- coding: utf-8 -*-

"""
    The predict module
    ======================

    Use it to predict some images from a trained network.
"""

import logging
import os
import time

import cv2
import numpy as np
import torch
from tqdm import tqdm

import doc_ufcn.train.utils.prediction as pr_utils


def get_predicted_polygons(
    probas: np.ndarray, min_cc: int, classes_names: list
) -> dict:
    """
    Clean the predicted and retrieve the detected object coordinates.
    :param probas: The probability maps obtained by the model.
    :param min_cc: The threshold used to remove small connected components.
    :param classes_names: The classes names involved in the experiment.
    :return page_contours: The contour and confidence score obtained for each
                           detected object.
    """
    page_contours = {}
    max_probas = np.argmax(probas, axis=0)
    for channel in range(1, probas.shape[0]):
        # Keep pixels with highest probability.
        channel_probas = np.uint8(max_probas == channel) * probas[channel, :, :]
        # Retrieve the polygons contours.
        bin_img = channel_probas.copy()
        bin_img[bin_img > 0] = 1
        contours, hierarchy = cv2.findContours(
            np.uint8(bin_img), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )
        # Remove small connected components.
        if min_cc > 0:
            contours = [
                contour for contour in contours if cv2.contourArea(contour) > min_cc
            ]
        page_contours[classes_names[channel]] = [
            {
                "confidence": pr_utils.compute_confidence(contour, channel_probas),
                "polygon": contour,
            }
            for contour in contours
        ]
    return page_contours


def run(
    prediction_path: str,
    log_path: str,
    img_size: int,
    colors: list,
    classes_names: list,
    save_image: list,
    min_cc: int,
    loaders: dict,
    net,
):
    """
    Run the prediction.
    :param prediction_path: The path to save the predictions.
    :param log_path: Path to save the experiment information and model.
    :param img_size: Network input size.
    :param colors: Colors of the classes used during the experiment.
    :param classes_names: The names of the classes involved during the experiment.
    :param save_image: List of sets (train, val, test) for which the prediction images
                       are generated and saved.
    :param min_cc: The threshold used to remove small connected components.
    :param loaders: The loaders containing the images to predict.
    :param net: The loaded network.
    """
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    # Run prediction.
    net.eval()

    logging.info("Starting predicting")
    starting_time = time.time()

    with torch.no_grad():
        for set, loader in zip(["train", "val", "test"], loaders.values()):
            seen_datasets = []
            # Create folders to save the predictions.
            os.makedirs(os.path.join(log_path, prediction_path, set), exist_ok=True)

            for i, data in enumerate(tqdm(loader, desc="Prediction (prog) " + set), 0):
                # Create dataset folders to save the predictions.
                if data["dataset"][0] not in seen_datasets:
                    os.makedirs(
                        os.path.join(
                            log_path, prediction_path, set, data["dataset"][0]
                        ),
                        exist_ok=True,
                    )
                    seen_datasets.append(data["dataset"][0])

                # Generate and save the predictions.
                output = net(data["image"].to(device).float())
                input_size = [element.numpy()[0] for element in data["size"][:2]]

                assert output.shape[0] == 1
                polygons = get_predicted_polygons(
                    output[0].cpu().numpy(), min_cc, classes_names
                )
                polygons = pr_utils.resize_polygons(
                    polygons, input_size, img_size, data["padding"]
                )

                polygons["img_size"] = [int(element) for element in input_size]
                pr_utils.save_prediction(
                    polygons,
                    os.path.join(
                        log_path,
                        prediction_path,
                        set,
                        data["dataset"][0],
                        data["name"][0],
                    ),
                )
                if set in save_image:
                    pr_utils.save_prediction_image(
                        polygons,
                        colors,
                        input_size,
                        os.path.join(
                            log_path,
                            prediction_path,
                            set,
                            data["dataset"][0],
                            data["name"][0],
                        ),
                    )

    end = time.gmtime(time.time() - starting_time)
    logging.info(
        "Finished predicting in %2d:%2d:%2d", end.tm_hour, end.tm_min, end.tm_sec
    )



# -*- coding: utf-8 -*-

"""
    The preprocessing module
    ======================

    Use it to preprocess the images.
"""

import os

import cv2
import numpy as np
import torch
from torch.utils.data import Dataset

from doc_ufcn.train.utils import rgb_to_gray_array, rgb_to_gray_value


class TrainingDataset(Dataset):
    """
    The TrainingDataset class is used to prepare the images and labels to
    run training step.
    """

    def __init__(
        self, images_dir: str, masks_dir: str, colors: list, transform: list = None
    ):
        """
        Constructor of the TrainingDataset class.
        :param images_dir: The directories containing the images.
        :param masks_dir: The directories containing the masks of the images.
        :param colors: The color codes of the different classes.
        :param transform: The list of the transformations to apply.
        """
        self.images_dir = images_dir
        self.images = [
            (dir.parent.parent.name, dir / element)
            for dir in self.images_dir
            for element in os.listdir(dir)
        ]
        self.masks_dir = masks_dir
        self.masks = {dir.parent.parent.name: dir for dir in self.masks_dir}
        self.colors = colors
        self.transform = transform

    def __len__(self) -> int:
        """
        Get the size of the dataset.
        :return: The size of the dataset.
        """
        return len(self.images)

    def __getitem__(self, idx: int) -> dict:
        """
        Get a sample from the dataset according to an index.
        :param idx: The index of the wanted sample.
        :return sample: The sample with index idx.
        """
        if torch.is_tensor(idx):
            idx = idx.tolist()
        img_name = self.images[idx]

        image = cv2.imread(str(img_name[1]))
        if len(image.shape) < 3:
            image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        label = cv2.imread(str(self.masks[img_name[0]] / img_name[1].name))
        if len(label.shape) < 3:
            label = cv2.cvtColor(label, cv2.COLOR_GRAY2RGB)
        label = cv2.cvtColor(label, cv2.COLOR_BGR2RGB)
        label = rgb_to_gray_array(label)

        # Transform the label into a categorical label.
        new_label = np.zeros_like(label)
        for index, value in enumerate(self.colors):
            color = rgb_to_gray_value(value)
            new_label[label == color] = index

        sample = {"image": image, "mask": new_label, "size": image.shape[0:2]}

        # Apply the transformations.
        if self.transform:
            sample = self.transform(sample)

        sample["size"] = sample["image"].shape[0:2]

        return sample


class PredictionDataset(Dataset):
    """
    The PredictionDataset class is used to prepare the images to
    run prediction step.
    """

    def __init__(self, images_dir: str, transform: list = None):
        """
        Constructor of the PredictionDataset class.
        :param images_dir: The directories containing the images.
        :param transform: The list of the transformations to apply.
        """
        self.images_dir = images_dir
        self.images = [
            (dir.parent.parent.name, dir / element)
            for dir in self.images_dir
            for element in os.listdir(dir)
        ]
        self.transform = transform

    def __len__(self) -> int:
        """
        Get the size of the dataset.
        :return: The size of the dataset.
        """
        return len(self.images)

    def __getitem__(self, idx: int) -> dict:
        """
        Get a sample from the dataset according to an index.
        :param idx: The index of the wanted sample.
        :return sample: The sample with index idx.
        """
        if torch.is_tensor(idx):
            idx = idx.tolist()
        img_name = self.images[idx][1]
        image = cv2.imread(str(img_name))

        if len(image.shape) < 3:
            image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        sample = {
            "image": image,
            "name": img_name.name,
            "dataset": self.images[idx][0],
            "size": image.shape[0:2],
        }

        # Apply the transformations.
        if self.transform:
            sample = self.transform(sample)
        return sample


# Transformations


class Rescale:
    """
    The Rescale class is used to rescale the image of a sample into a
    given size.
    """

    def __init__(self, output_size: int):
        """
        Constructor of the Rescale class.
        :param output_size: The desired new size.
        """
        assert isinstance(output_size, int)
        self.output_size = output_size

    def __call__(self, sample: dict) -> dict:
        """
        Rescale the sample image into the model input size.
        :param sample: The sample to rescale.
        :return sample: The rescaled sample.
        """
        old_size = sample["image"].shape[:2]
        # Compute the new sizes.
        ratio = float(self.output_size) / max(old_size)
        new_size = [int(x * ratio) for x in old_size]

        # Resize the image.
        if max(old_size) != self.output_size:
            image = cv2.resize(sample["image"], (new_size[1], new_size[0]))
            sample["image"] = image

        # Resize the label. MUST BE AVOIDED.
        if "mask" in sample.keys():
            if max(sample["mask"].shape[:2]) != self.output_size:
                mask = cv2.resize(sample["mask"], (new_size[1], new_size[0]))
                sample["mask"] = mask
        return sample


class Pad:
    """
    The Pad class is used to pad the image of a sample to make it divisible by 8.
    """

    def __init__(self):
        """
        Constructor of the Pad class.
        """
        pass

    def __call__(self, sample: dict) -> dict:
        """
        Pad the sample image with zeros.
        :param sample: The sample to pad.
        :return sample: The padded sample.
        """
        # Compute the padding parameters.
        delta_w = 0
        delta_h = 0
        if sample["image"].shape[0] % 8 != 0:
            delta_h = (
                int(8 * np.ceil(sample["image"].shape[0] / 8))
                - sample["image"].shape[0]
            )
        if sample["image"].shape[1] % 8 != 0:
            delta_w = (
                int(8 * np.ceil(sample["image"].shape[1] / 8))
                - sample["image"].shape[1]
            )

        top, bottom = delta_h // 2, delta_h - (delta_h // 2)
        left, right = delta_w // 2, delta_w - (delta_w // 2)

        # Add padding to have same size images.
        image = cv2.copyMakeBorder(
            sample["image"],
            top,
            bottom,
            left,
            right,
            cv2.BORDER_CONSTANT,
            value=[0, 0, 0],
        )
        sample["image"] = image
        sample["padding"] = {"top": top, "left": left}
        return sample


class Normalize:
    """
    The Normalize class is used to normalize the image of a sample.
    The mean value and standard deviation must be first computed on the
    training dataset.
    """

    def __init__(self, mean: list, std: list):
        """
        Constructor of the Normalize class.
        :param mean: The mean values (one for each channel) of the images
                     pixels of the training dataset.
        :param std: The standard deviations (one for each channel) of the
                    images pixels of the training dataset.
        """
        assert isinstance(mean, list)
        assert isinstance(std, list)
        self.mean = mean
        self.std = std

    def __call__(self, sample: dict) -> dict:
        """
        Normalize the sample image.
        :param sample: The sample with the image to normalize.
        :return sample: The sample with the normalized image.
        """
        image = np.zeros(sample["image"].shape)
        for channel in range(sample["image"].shape[2]):
            image[:, :, channel] = (
                np.float32(sample["image"][:, :, channel]) - self.mean[channel]
            ) / self.std[channel]
        sample["image"] = image
        return sample


class ToTensor:
    """
    The ToTensor class is used convert ndarrays into Tensors.
    """

    def __call__(self, sample: dict) -> dict:
        """
        Transform the sample image and label into Tensors.
        :param sample: The initial sample.
        :return sample: The sample made of Tensors.
        """
        sample["image"] = torch.from_numpy(sample["image"].transpose((2, 0, 1)))
        if "mask" in sample.keys():
            sample["mask"] = torch.from_numpy(sample["mask"])
        return sample



# -*- coding: utf-8 -*-
import hashlib
from pathlib import Path


def md5sum(path):
    """
    Calc the MD5 hash of a binary file
    """
    with open(path, mode="rb") as f:
        d = hashlib.md5()
        while True:
            buf = f.read(4096)
            if not buf:
                break
            d.update(buf)
        return d.hexdigest()


def export_list(data: list, output: Path):
    """
    Export a list of elements to a specified location, one element per line.
    """
    with output.open("w") as writer:
        writer.write("\n".join(str(value) for value in data))



# -*- coding: utf-8 -*-

import copy
import logging
import os
import sys
import time

import torch
from torch.cuda.amp import autocast
from torch.nn import Module as NNModule

logger = logging.getLogger(__name__)


class DocUFCNModel(NNModule):
    """
    The DocUFCNModel class is used to generate the Doc-UFCN network.
    The class initializes different useful layers and defines
    the sequencing of the defined layers/blocks.
    """

    def __init__(self, no_of_classes, use_amp=False):
        """
        Constructor of the DocUFCNModel class.
        :param no_of_classes: The number of classes wanted at the
                              output of the network.
        :param use_amp: Whether to use Automatic Mixed Precision.
                        Disabled by default
        """
        super(DocUFCNModel, self).__init__()
        self.amp = use_amp
        self.dilated_block1 = self.dilated_block(3, 32)
        self.dilated_block2 = self.dilated_block(32, 64)
        self.dilated_block3 = self.dilated_block(64, 128)
        self.dilated_block4 = self.dilated_block(128, 256)
        self.pool = torch.nn.MaxPool2d(2, 2)
        self.conv_block1 = self.conv_block(256, 128)
        self.conv_block2 = self.conv_block(256, 64)
        self.conv_block3 = self.conv_block(128, 32)
        self.last_conv = torch.nn.Conv2d(64, no_of_classes, 3, stride=1, padding=1)
        self.softmax = torch.nn.Softmax(dim=1)

    @staticmethod
    def dilated_block(input_size, output_size):
        """
        Define a dilated block.
        It consists in 6 successive convolutions with the dilations
        rates [1, 2, 4, 8, 16].
        :param input_size: The size of the input tensor.
        :param output_size: The size of the output tensor.
        :return: The sequence of the convolutions.
        """
        modules = []
        modules.append(
            torch.nn.Conv2d(
                input_size, output_size, 3, stride=1, dilation=1, padding=1, bias=False
            )
        )
        modules.append(torch.nn.BatchNorm2d(output_size, track_running_stats=False))
        modules.append(torch.nn.ReLU(inplace=True))
        modules.append(torch.nn.Dropout(p=0.4))
        for i in [2, 4, 8, 16]:
            modules.append(
                torch.nn.Conv2d(
                    output_size,
                    output_size,
                    3,
                    stride=1,
                    dilation=i,
                    padding=i,
                    bias=False,
                )
            )
            modules.append(torch.nn.BatchNorm2d(output_size, track_running_stats=False))
            modules.append(torch.nn.ReLU(inplace=True))
            modules.append(torch.nn.Dropout(p=0.4))
        return torch.nn.Sequential(*modules)

    @staticmethod
    def conv_block(input_size, output_size):
        """
        Define a convolutional block.
        It consists in a convolution followed by an upsampling layer.
        :param input_size: The size of the input tensor.
        :param output_size: The size of the output tensor.
        :return: The sequence of the convolutions.
        """
        return torch.nn.Sequential(
            torch.nn.Conv2d(
                input_size, output_size, 3, stride=1, padding=1, bias=False
            ),
            torch.nn.BatchNorm2d(output_size, track_running_stats=False),
            torch.nn.ReLU(inplace=True),
            torch.nn.Dropout(p=0.4),
            # Does the upsampling.
            torch.nn.ConvTranspose2d(output_size, output_size, 2, stride=2, bias=False),
            torch.nn.BatchNorm2d(output_size, track_running_stats=False),
            torch.nn.ReLU(inplace=True),
            torch.nn.Dropout(p=0.4),
        )

    def forward(self, input_tensor):
        """
        Define the forward step of the network.
        It consists in 4 successive dilated blocks followed by 3
        convolutional blocks, a final convolution and a softmax layer.
        :param input_tensor: The input tensor.
        :return: The output tensor.
        """
        with autocast(enabled=self.amp):
            tensor = self.dilated_block1(input_tensor)
            out_block1 = tensor
            tensor = self.dilated_block2(self.pool(tensor))
            out_block2 = tensor
            tensor = self.dilated_block3(self.pool(tensor))
            out_block3 = tensor
            tensor = self.dilated_block4(self.pool(tensor))
            tensor = self.conv_block1(tensor)
            tensor = torch.cat([tensor, out_block3], dim=1)
            tensor = self.conv_block2(tensor)
            tensor = torch.cat([tensor, out_block2], dim=1)
            tensor = self.conv_block3(tensor)
            tensor = torch.cat([tensor, out_block1], dim=1)
            output_tensor = self.last_conv(tensor)
            return self.softmax(output_tensor)


def weights_init(model):
    """
    Initialize the model weights.
    :param model: The model.
    """
    if isinstance(model, (torch.nn.Conv2d, torch.nn.ConvTranspose2d)):
        torch.nn.init.xavier_uniform_(model.weight.data)


def load_network(no_of_classes: int, use_amp: bool):
    """
    Load the network for the experiment.
    :param no_of_classes: The number of classes involved in the experiment.
    :param use_amp: Whether to use Automatic Mixed Precision.
    :return net: The loaded network.
    :return last_layer: The last activation function to apply.
    """
    # Define the network.
    net = DocUFCNModel(no_of_classes, use_amp)
    # Allow parallel running if more than 1 gpu available.
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    logger.info("Running on %s", device)
    if torch.cuda.device_count() > 1:
        logger.info("Let's use %d GPUs", torch.cuda.device_count())
        net = torch.nn.DataParallel(net)
    return net.to(device)


def restore_model(
    net, optimizer, scaler, log_path: str, model_path: str, keep_last: bool = True
):
    """
    Load the model weights.
    :param net: The loaded model.
    :param optimizer: The loaded optimizer.
    :param scaler: The scaler used for AMP.
    :param log_path: The directory containing the model to restore.
    :param model_path: The name of the model to restore.
    :param keep_last: Retrieve the last layer weights.
    :return checkpoint: The loaded checkpoint.
    :return net: The restored model.
    :return optimizer: The restored optimizer.
    :return scaler: The restored scaler.
    """
    starting_time = time.time()
    if not os.path.isfile(os.path.join(log_path, model_path)):
        logger.error("No model found at %s", os.path.join(log_path, model_path))
        sys.exit()
    else:
        if torch.cuda.is_available():
            checkpoint = torch.load(os.path.join(log_path, model_path))
        else:
            checkpoint = torch.load(
                os.path.join(log_path, model_path), map_location=torch.device("cpu")
            )
        loaded_checkpoint = {}
        if torch.cuda.device_count() > 1:
            for key in checkpoint["state_dict"].keys():
                if "module" not in key:
                    loaded_checkpoint["module." + key] = checkpoint["state_dict"][key]
                else:
                    loaded_checkpoint = checkpoint["state_dict"]
        else:
            for key in checkpoint["state_dict"].keys():
                loaded_checkpoint[key.replace("module.", "")] = checkpoint[
                    "state_dict"
                ][key]

        if keep_last:
            net.load_state_dict(loaded_checkpoint)
            if optimizer is not None:
                optimizer.load_state_dict(checkpoint["optimizer"])
            if scaler is not None:
                scaler.load_state_dict(checkpoint["scaler"])
        else:
            loaded_checkpoint.pop("last_conv.weight")
            loaded_checkpoint.pop("last_conv.bias")
            net.load_state_dict(loaded_checkpoint, strict=False)

        logger.info(
            "Loaded checkpoint %s (epoch %d) in %1.5fs",
            model_path,
            checkpoint["epoch"],
            (time.time() - starting_time),
        )
        return checkpoint, net, optimizer, scaler


def save_model(epoch: int, model, loss: float, optimizer, scaler, filename: str):
    """
    Save the given model.
    :param epoch: The current epoch.
    :param model: The model state dict to save.
    :param loss: The loss of the current epoch.
    :param optimizer: The optimizer state dict.
    :param scaler: The scaler used for AMP.
    :param filename: The name of the model file.
    """
    model_params = {
        "epoch": epoch,
        "state_dict": copy.deepcopy(model),
        "best_loss": loss,
        "optimizer": copy.deepcopy(optimizer),
        "scaler": scaler,
    }
    torch.save(model_params, filename)



# -*- coding: utf-8 -*-

from pathlib import Path

from teklia_toolbox.config import ConfigParser


def parse_configurations(config_path: Path):
    """
    Parse multiple YAML configuration files into a single source
    of configuration for the HuggingFace app

    :param config_path: pathlib.Path, Path to the .yaml config file
    :return: dict, containing the configuration. Ensures config is complete and with correct typing
    """
    parser = ConfigParser()

    parser.add_option("title")
    parser.add_option("description")
    parser.add_option("examples", type=list)
    model_parser = parser.add_subparser("models", many=True)

    model_parser.add_option("model_name")
    model_parser.add_option("title")
    model_parser.add_option("description")
    model_parser.add_option("classes_colors", type=list)

    return parser.parse(config_path)



# -*- coding: utf-8 -*-

"""
    The normalization params module
    ======================

    Use it to get the mean and standard deviation of the training set.
"""

import logging

import numpy as np
from torch.utils.data import DataLoader
from torchvision import transforms
from tqdm import tqdm

from doc_ufcn.train.utils.preprocessing import PredictionDataset, Rescale, ToTensor


def run(
    log_path: str,
    data_paths: dict,
    img_size: int,
    mean_name: str,
    std_name: str,
    num_workers: int = 2,
):
    """
    Compute the normalization parameters: mean and standard deviation on
    train set.
    :param log_path: Path to save the experiment information and model.
    :param data_paths: Path to the data folders.
    :param img_size: The network input image size.
    :param mean_name: Name of the file that will contain all the mean values.
    :param std_name: Name of the file that will contain all the std values.
    """
    dataset = PredictionDataset(
        data_paths["train"]["image"],
        transform=transforms.Compose([Rescale(img_size), ToTensor()]),
    )
    loader = DataLoader(dataset, batch_size=1, shuffle=False, num_workers=num_workers)

    # Compute mean and std.
    mean = []
    std = []
    for data in tqdm(loader, desc="Computing parameters (prog)"):
        image = data["image"].numpy()
        mean.append(np.mean(image, axis=(0, 2, 3)))
        std.append(np.std(image, axis=(0, 2, 3)))

    mean = np.array(mean).mean(axis=0)
    std = np.array(std).mean(axis=0)

    logging.info("Mean: {}".format(np.uint8(mean)))
    logging.info(" Std: {}".format(np.uint8(std)))

    with (log_path / mean_name).open("w") as file:
        for value in mean:
            file.write(str(np.uint8(value)) + "\n")

    with (log_path / std_name).open("w") as file:
        for value in std:
            file.write(str(np.uint8(value)) + "\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path

from setuptools import find_packages, setup


def parse_requirements(name):
    path = Path(__file__).parent.resolve() / name
    assert path.exists(), f"Missing requirements: {path}"
    return list(map(str.strip, path.read_text().splitlines()))


setup(
    name="doc-ufcn",
    version=open("VERSION").read(),
    description="Doc-UFCN",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Mélodie Boillet",
    author_email="boillet@teklia.com",
    project_urls={
        "Source": "https://gitlab.teklia.com/dla/doc-ufcn/",
        "Tracker": "https://gitlab.teklia.com/dla/doc-ufcn/issues/",
    },
    url="https://gitlab.com/teklia/dla/doc-ufcn",
    install_requires=parse_requirements("requirements.txt"),
    extras_require={
        "training": parse_requirements("training-requirements.txt"),
    },
    packages=find_packages(exclude=["tests"]),
    python_requires=">= 3.8, < 3.11",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        # Specify the Python versions you support here.
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        # Topics
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
    ],
)



# -*- coding: utf-8 -*-
import json
import os
from pathlib import Path

from teklia_toolbox.config import ConfigParser, ConfigurationError

from doc_ufcn.train import logger

STEPS = ["normalization_params", "train", "prediction", "evaluation"]


def parse_configurations(paths):
    """
    Parse multiple JSON configuration files into a single source
    of configuration for the whole training workflow
    """

    def _step(value: str):
        if value not in STEPS:
            raise ConfigurationError(f"Invalid step {value}")
        return value

    def _user_path(value: str):
        return Path(value).expanduser()

    def _loss(value: str):
        value = value.lower()
        if value not in ["initial", "best"]:
            raise ConfigurationError(f"Invalid loss {value}")
        return value

    def _same_classes(value: str):
        value = value.lower()
        if value not in ["true", "false"]:
            raise ConfigurationError(
                f"Invalid same classes argument: {value}. This value should be set to True or False"
            )
        return value == "true"

    def _rgb(value: list):
        if not isinstance(value, (list, tuple)):
            raise ConfigurationError("This RGB value should be a list or tuple")

        if len(value) != 3:
            raise ConfigurationError("This RGB value should be a list with 3 items")

        if not all(isinstance(v, int) for v in value):
            raise ConfigurationError("This RGB value should be set with integers only")

        if not all(0 <= v <= 255 for v in value):
            raise ConfigurationError(
                "This RGB value should be set with integers in range 0-255"
            )

        return value

    parser = ConfigParser()
    parser.add_option("experiment_name", type=str, default="doc-ufcn")

    # List of the steps to run.
    parser.add_option("steps", type=_step, many=True, default=STEPS)

    # Global parameters of the experiment entered by the user.
    parser.add_option(
        "classes_names", type=str, many=True, default=["background", "text_line"]
    )
    parser.add_option(
        "classes_colors", type=_rgb, many=True, default=[[0, 0, 0], [0, 0, 255]]
    )
    parser.add_option("img_size", type=int, default=768)
    parser.add_option("no_of_epochs", type=int, default=100)
    parser.add_option("batch_size", type=int, default=None)
    parser.add_option("no_of_params", type=int, default=None)
    parser.add_option("bin_size", type=int, default=20)
    parser.add_option("learning_rate", type=float, default=5e-3)
    parser.add_option("min_cc", type=int, default=0)
    parser.add_option("save_image", type=str, many=True, default=[])
    parser.add_option("use_amp", type=bool, default=False)
    parser.add_option("mean", type=str, default="mean")
    parser.add_option("std", type=str, default="std")
    parser.add_option("model_path", type=Path, default=Path("model.pth"))
    parser.add_option("prediction_path", type=Path, default=Path("prediction"))
    parser.add_option("evaluation_path", type=Path, default=Path("evaluation"))

    # Path to save the Tensorboard events.
    parser.add_option("tb_path", type=Path, default=Path("events"))

    # Path to save the experiment information and model.
    parser.add_option("log_path", type=Path, default=None)

    # Path to the data folders.
    data_paths = parser.add_subparser("data_paths")

    train_paths = data_paths.add_subparser("train")
    train_paths.add_option(
        "image", type=_user_path, default=Path("./data/train/images"), many=True
    )
    train_paths.add_option(
        "mask", type=_user_path, default=Path("./data/train/labels"), many=True
    )
    train_paths.add_option(
        "json", type=_user_path, default=Path("./data/train/labels_json"), many=True
    )

    val_paths = data_paths.add_subparser("val")
    val_paths.add_option(
        "image", type=_user_path, default=Path("./data/val/images"), many=True
    )
    val_paths.add_option(
        "mask", type=_user_path, default=Path("./data/val/labels"), many=True
    )
    val_paths.add_option(
        "json", type=_user_path, default=Path("./data/val/labels_json"), many=True
    )

    test_paths = data_paths.add_subparser("test")
    test_paths.add_option(
        "image", type=_user_path, default=Path("./data/test/images"), many=True
    )
    test_paths.add_option(
        "json", type=_user_path, default=Path("./data/test/labels_json"), many=True
    )

    # Training parameters.
    training = parser.add_subparser("training", default={})
    training.add_option("restore_model", type=str, default=None)
    training.add_option("same_classes", type=_same_classes, default="true")
    training.add_option("loss", type=_loss, default="initial")

    # MLflow parameters
    mlflow = parser.add_subparser("mlflow", default=None)
    mlflow.add_option("experiment_id", type=str)
    mlflow.add_option("tracking_uri", type=str)
    mlflow.add_option("s3_endpoint_url", type=str)
    mlflow.add_option("aws_access_key_id", type=str, default=None)
    mlflow.add_option("aws_secret_access_key", type=str, default=None)
    mlflow.add_option("run_name", type=str, default=None)

    # Merge all provided configuration files into a single payload
    # that will be validated by the configuration parser described above
    raw = {}
    for path in paths:
        try:
            raw.update(json.load(path.open()))
        except Exception as e:
            logger.error(f"Failed to parse config {path} : {e}")
            raise Exception("Invalid configuration")

    # Promote deprecated parameters to root level
    for deprecated_key in ("params", "global_params"):
        if deprecated_key in raw:
            logger.warn(
                f"Promoting {deprecated_key} to root configuration level. You should update your configuration to promote the parameters directly."
            )
            deprecated = raw.pop(deprecated_key)
            raw.update(deprecated)

    out = parser.parse_data(raw)

    assert (
        out["batch_size"] is not None or out["no_of_params"] is not None
    ), "Please provide a batch size or a maximum number of parameters"

    # Update log path using experiment name
    if out["log_path"] is None:
        slug = out["experiment_name"].lower().replace(" ", "_").replace("-", "_")
        out["log_path"] = Path("./runs") / slug

    return out


def save_configuration(config: dict):
    """
    Save the current configuration.
    :param log_path: Path to save the experiment information and model.
    :param experiment_name: The name of the experiment that is used to save all
                      the experiment information.
    :param config : Full configuration payload that will be saved and usable to retry the experiment
    """
    os.makedirs(config["log_path"], exist_ok=True)
    path = config["log_path"] / (config["experiment_name"] + ".json")
    with open(path, "w") as config_file:
        json.dump(config, config_file, indent=4, default=str, sort_keys=True)
        logger.info(f"Saved configuration in {path.resolve()}")



# -*- coding: utf-8 -*-

"""
    The evaluation module
    ======================

    Use it to evaluation a trained network.
"""

import logging
import os
import time
from pathlib import Path

import mlflow
import numpy as np
from tqdm import tqdm

import doc_ufcn.train.utils.evaluation as ev_utils
import doc_ufcn.train.utils.object_metrics as o_metrics
import doc_ufcn.train.utils.pixel_metrics as p_metrics


def run(
    log_path: str,
    classes_names: list,
    set: str,
    data_paths: dict,
    dataset: str,
    prediction_path: Path,
    evaluation_path: Path,
    mlflow_logging: bool,
):
    """
    Run the evaluation.
    :param log_path: Path to save the evaluation results and load the model.
    :param classes_names: The names of the classes involved during the experiment.
    :param set: The current evaluated set.
    :param data_paths: Path to the data folders.
    :param dataset: The dataset to evaluate.
    :param prediction_path: Path where the prediction has been written.
    :param evaluation_path: Path where the evaluation will been written.
    :param mlflow_logging: Whether we should log data to MLflow.
    """
    # Run evaluation.
    logging.info("Starting evaluation: " + dataset)
    starting_time = time.time()

    label_dir = [dir for dir in data_paths if dataset in str(dir)][0]

    pixel_metrics = {
        channel: {metric: [] for metric in ["iou", "precision", "recall", "fscore"]}
        for channel in classes_names[1:]
    }
    object_metrics = {
        channel: {metric: {} for metric in ["precision", "recall", "fscore", "AP"]}
        for channel in classes_names[1:]
    }
    rank_scores = {
        channel: {
            iou: {rank: {"True": 0, "Total": 0} for rank in range(95, -5, -5)}
            for iou in range(50, 100, 5)
        }
        for channel in classes_names[1:]
    }
    number_of_gt = {channel: 0 for channel in classes_names[1:]}
    for img_name in tqdm(os.listdir(label_dir), desc="Evaluation (prog) " + set):
        gt_regions = ev_utils.read_json(os.path.join(label_dir, img_name))
        pred_regions = ev_utils.read_json(
            os.path.join(log_path, prediction_path, set, dataset, img_name)
        )

        gt_polys = ev_utils.get_polygons(gt_regions, classes_names)

        if gt_regions["img_size"] != pred_regions["img_size"]:
            pred_regions = ev_utils.resize_polygons(
                pred_regions, gt_regions["img_size"], pred_regions["img_size"]
            )
        pred_polys = ev_utils.get_polygons(pred_regions, classes_names)

        pixel_metrics = p_metrics.compute_metrics(
            gt_polys, pred_polys, classes_names[1:], pixel_metrics
        )

        image_rank_scores = o_metrics.compute_rank_scores(
            gt_polys, pred_polys, classes_names[1:]
        )
        rank_scores = o_metrics.update_rank_scores(
            rank_scores, image_rank_scores, classes_names[1:]
        )
        number_of_gt = {
            channel: number_of_gt[channel] + len(gt_polys[channel])
            for channel in classes_names[1:]
        }

    object_metrics = o_metrics.get_mean_results(
        rank_scores, number_of_gt, classes_names[1:], object_metrics
    )

    if mlflow_logging:
        # Log metrics per channel
        for channel in classes_names[1:]:
            prefix = f"{set.upper()} {channel}"
            # Pixel metrics
            prefixed_pixel_metrics = {
                f"{prefix}_{metric}": np.round(np.mean(value), 4)
                for metric, value in pixel_metrics[channel].items()
            }
            aps = object_metrics[channel]["AP"]
            # AP values
            AP_metrics = {
                f"{prefix} AP IOU_0.{level}": np.round(aps[level], 4)
                for level in [50, 75, 95]
            }
            AP_metrics[f"{prefix} AP_0.5-0.95"] = np.round(
                np.mean(list(aps.values())), 4
            )
            mlflow.set_tags(tags={**prefixed_pixel_metrics, **AP_metrics})

    # Print the results.
    print(set)
    for channel in classes_names[1:]:
        print(channel)
        print("IoU       = ", np.round(np.mean(pixel_metrics[channel]["iou"]), 4))
        print("Precision = ", np.round(np.mean(pixel_metrics[channel]["precision"]), 4))
        print("Recall    = ", np.round(np.mean(pixel_metrics[channel]["recall"]), 4))
        print("F-score   = ", np.round(np.mean(pixel_metrics[channel]["fscore"]), 4))

        aps = object_metrics[channel]["AP"]
        print("AP [IOU=0.50] = ", np.round(aps[50], 4))
        print("AP [IOU=0.75] = ", np.round(aps[75], 4))
        print("AP [IOU=0.95] = ", np.round(aps[95], 4))
        print("AP [0.5,0.95] = ", np.round(np.mean(list(aps.values())), 4))
        print("\n")

    os.makedirs(os.path.join(log_path, evaluation_path, set), exist_ok=True)
    # ev_utils.save_graphical_results(object_metrics, classes_names[1:],
    #                                os.path.join(log_path, params.evaluation_path, set))
    ev_utils.save_results(
        pixel_metrics,
        object_metrics,
        classes_names[1:],
        os.path.join(log_path, evaluation_path, set),
        dataset,
    )

    end = time.gmtime(time.time() - starting_time)
    logging.info(
        "Finished evaluating in %2d:%2d:%2d", end.tm_hour, end.tm_min, end.tm_sec
    )



# -*- coding: utf-8 -*-

from dataclasses import dataclass, field

from doc_ufcn import models
from doc_ufcn.main import DocUFCN


@dataclass
class UFCNModel:
    name: str
    colors: list
    title: str
    description: str
    classes: list = field(default_factory=list)
    model: DocUFCN = None

    def get_class_name(self, channel_idx):
        return self.classes[channel_idx]

    @property
    def loaded(self):
        return self.model is not None

    @property
    def num_channels(self):
        return len(self.classes)

    def load(self):
        # Download the model
        model_path, parameters = models.download_model(name=self.name)

        # Store classes
        self.classes = parameters["classes"]

        # Check that the number of colors is equal to the number of classes -1
        assert self.num_channels - 1 == len(
            self.colors
        ), f"The parameter classes_colors was filled with the wrong number of colors. {self.num_channels-1} colors are expected instead of {len(self.colors)}."

        # Load the model
        self.model = DocUFCN(
            no_of_classes=len(self.classes),
            model_input_size=parameters["input_size"],
            device="cpu",
        )
        self.model.load(
            model_path=model_path, mean=parameters["mean"], std=parameters["std"]
        )



# -*- coding: utf-8 -*-
import argparse
import logging
from pathlib import Path

from doc_ufcn.train.configuration import parse_configurations, save_configuration
from doc_ufcn.train.experiment import run

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


logger = logging.getLogger(__name__)


def parse_args():
    """
    Configure the CLI arguments for the training workflow
    """
    parser = argparse.ArgumentParser(description="Run a Doc-UFCN training experiment")
    parser.add_argument(
        "config",
        type=Path,
        nargs="+",
        help="Path towards an experiment or global configuration file. All files will be merged in order.",
    )

    return parser.parse_args()


def main():
    args = parse_args()

    # Parse all configuration files provided from CLI
    config = parse_configurations(args.config)

    # Ensure we have at least one step to run
    if len(config["steps"]) == 0:
        logger.error("No step to run, exiting execution.")
        return

    # Save configuration to be able to re-run the experiment
    save_configuration(config)

    # Run experiment
    run(config)


if __name__ == "__main__":
    main()
